from tests.fixtures.regressors.simple_gaussian_cnn_regressor import (
    SimpleGaussianCNNRegressor)
from tests.fixtures.regressors.simple_gaussian_mlp_regressor import (
    SimpleGaussianMLPRegressor)
from tests.fixtures.regressors.simple_mlp_regressor import (SimpleMLPRegressor)

__all__ = [
    'SimpleGaussianCNNRegressor', 'SimpleGaussianMLPRegressor',
    'SimpleMLPRegressor'
]
